/*!
 * Copyright (C) 2001-2003 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Log$
 */

#include <dev/board.h>
#include <dev/debug.h>
#include <dev/urom.h>

#include <sys/version.h>
#include <sys/atom.h>
#include <sys/heap.h>
#include <sys/timer.h>
#include <sys/device.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include <hxmp3/mp3dec.h>
#include <hxmp3/hermite.h>

#include "at73dac.h"

#define MAX_FILESIZE    (1024 * 1024)

#define BASE_SAMPLE_RATE    44100

HMP3Decoder hdec;
MP3FrameInfo frameinfo;

void *hres;
short pcmdec[4096];
short pcmbuf[9216];

void At73DacDump(void)
{
    u_int reg;

    for(reg = 0x00; reg < 0x80; reg++) {
        if ((reg & 0x0F) == 0x00) {
            putchar('\n');
        }
        printf("%02X:%02X ", reg, At73DacReadReg(reg));
    }
    putchar('\n');
}

/*
 * Play MP3 file from local file system.
 *
 * \param path Pathname of the MP3 file to play.
 *
 * \return 0 on success, -1 if opening the file failed.
 */
static int PlayMp3File(char *path)
{
    int fd;
    u_char *mp3buf;
    u_char *mp3ptr;
    int got;
    int skip;
    int rbytes;
    int ec;
    int first = 1;
    int os;

    /*
     * Open the MP3 file.
     */
    printf("Play %s: ", path);
    if ((fd = _open(path, _O_RDONLY | _O_BINARY)) == -1) {
        printf("Error %d\n", errno);
        return -1;
    }
    puts("OK");

    mp3buf = malloc(MAX_FILESIZE);
    got = _read(fd, mp3buf, MAX_FILESIZE);
    printf("[FileSize %d]", got);
    _close(fd);

    mp3ptr = mp3buf;
    rbytes = got;
    for (;;) {
        if (rbytes <= 0) {
            At73DacFlush();
            break;
        }
        if ((skip = MP3FindSyncWord(mp3ptr, rbytes)) < 0) {
            printf("\n[ERR-SYN]");
            break;
        }
        if (skip) {
            printf("[Skip %d]", skip);
            mp3ptr += skip;
            rbytes -= skip;
            if (rbytes <= 0) {
                At73DacFlush();
                break;
            }
        }
        if (first) {
            first = 0;
            ec = MP3GetNextFrameInfo(hdec, &frameinfo, mp3ptr);
            if (ec) {
                printf("\n[ERR-FRI %d]", ec);
                break;
            }
            printf("\nLayer %d, ", frameinfo.layer);
            printf("Version %d, ", frameinfo.version);
            printf("Bitrate %d, ", frameinfo.bitrate);
            printf("Channels %d, ", frameinfo.nChans);
            printf("Sample rate %d, ", frameinfo.samprate);
            printf("Bits/sample %d\n", frameinfo.bitsPerSample);
            if (frameinfo.nChans == 1) {
                frameinfo.samprate /= 2;
            }
            if ((hres = RAInitResamplerHermite(frameinfo.samprate, BASE_SAMPLE_RATE, frameinfo.nChans)) == NULL) {
                puts("Failed to init resampler.");
                break;
            }
        }

        ec = MP3Decode(hdec, &mp3ptr, &rbytes, pcmdec, 0);
        if (ec) {
            printf("\n[ERR-DEC %d]", ec);
            continue;
        }
        //printf("[%d]", frameinfo.outputSamps);
        putchar('.');
        if (frameinfo.nChans == 1) {
            os = RAResampleMonoHermite(pcmdec, frameinfo.outputSamps, pcmbuf, hres);
        }
        else {
            os = RAResampleStereoHermite(pcmdec, frameinfo.outputSamps, pcmbuf, hres);
        }
        if (At73DacWrite(pcmbuf, os)) {
            printf("\n[ERR-ADC]");
        }
    }
    if (hres) {
        RAFreeResamplerHermite(hres);
    }

    free(mp3buf);
    printf("\nDone, %u bytes free\n\n\n", (u_int)NutHeapAvailable());
    return 0;
}

/*!
 * \brief Main application routine. 
 *
 */
int main(void)
{
    /* Baudrate for debug output. */
    u_long baud = 115200;

    /*
     * Assign stdout to the debug device.
     */
    NutRegisterDevice(&DEV_DEBUG, 0, 0);
    freopen(DEV_DEBUG_NAME, "w", stdout);
    _ioctl(_fileno(stdout), UART_SETSPEED, &baud);
    printf("\n\nMP3 player on Nut/OS %s\n", NutVersionString());

    /*
     * Initialize file system.
     */
    NutRegisterDevice(&devUrom, 0, 0);

    printf("CPU Clock   : %lu\n", NutGetCpuClock());
    printf("Master Clock: %lu\n", At91GetMasterClock());
    printf("%lu system ticks/s\n", NutGetTickClock());

    {
        volatile long cntr = 10000000;
        u_long ticks = NutGetTickCount();
        while(cntr--);
        ticks = NutGetTickCount() - ticks;
        printf("%lu ticks: %lu MIPS\n", ticks, 60000 / ticks);
    }

    At73DacInit(BASE_SAMPLE_RATE);

    for (;;) {
        hdec = MP3InitDecoder();
        PlayMp3File("UROM:128kbps_44100Hz.mp3");
        MP3FreeDecoder(hdec);
        NutSleep(1000);

        hdec = MP3InitDecoder();
        PlayMp3File("UROM:056kbps_22050Hz_mono.mp3");
        MP3FreeDecoder(hdec);
        NutSleep(1000);

        hdec = MP3InitDecoder();
        PlayMp3File("UROM:032kbps_11025Hz.mp3");
        MP3FreeDecoder(hdec);
        NutSleep(1000);

        hdec = MP3InitDecoder();
        PlayMp3File("UROM:048kbps_16000Hz.mp3");
        MP3FreeDecoder(hdec);
        NutSleep(1000);

        hdec = MP3InitDecoder();
        PlayMp3File("UROM:056kbps_22050Hz.mp3");
        MP3FreeDecoder(hdec);
        NutSleep(1000);
    }
    for (;;);
}
